{-

  APCC
     
  Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015 
  Updated by Omar Almousa 2015

-}

module TypeSystem_OS where
import TypeSystem_Context
import Ast
import Msg
import TypeSystem_Defs
import TypeSystem_Evaluator
import StrandAST
import GeneratePlainStrands (generateStrands)
import OperationalStrands (opStrands)
import Data.List

-- APS protocol

-- build the initial context/environment for typing
buildContext :: Protocol -> Context
buildContext prot@(name, types, mappings, formats, knowledge, macros, actions, goals, private, results, abstraction) = 
                                let
                                    c1 = types2context defaultContext types
                                    c2 = mappings2context c1 mappings
                                    c3 = formats2context c2 formats
                                    c4 = actions2context c3 actions
                                in c4

-- compute secrecy goals from Private section
protprivate2goals :: Protocol-> Protocol
protprivate2goals prot@(name, types, mappings, formats, knowledge, macros, actions, goals, private, results, abstraction) =
                   (name, types,  mappings, formats, knowledge, macros, actions, goals1, private, results, abstraction) 
                              where 
                                    goals1 = nub(goals ++ private2goals private agents)
                                    agents = getAgentIds types

-- build goals checking messages structure
private2goals :: Private -> [Ident] -> Goals                                    
private2goals [] _ = []
private2goals _ [] = error "no agents specified"
private2goals (x:xs) agents = let
                                    idents = getIdentsofMsg [x]
                                    vars = filter isVariable idents
                                    agentsecret = intersect agents idents
                                    -- only variables of type agents are allowed
                                    goal = if null (vars \\ agents) then secGoal x agentsecret else error ("Private section messages should contain only variables of type Agents"        
                                                                                                                  ++ "\n\tvars: " ++ show vars
                                                                                                                  ++ "\n\tidents: " ++ show idents
                                                                                                                  ++ "\n\tagents: " ++ show agents
                                                                                                                  ++ "\n\tagentsecret: " ++ show agentsecret
                                                                                                                  ++ "\n\tmsg: " ++ show x
                                                                                                                  ++ "\n\tvar-agents:" ++ show (vars\\agents)
                                                                                                                  )
                              in goal : private2goals xs agents

secGoal :: Msg -> [Ident] -> Goal
secGoal msg [] = error ("no peers specfied in goal for msg: " ++ show msg)
secGoal msg (x:[]) = Secret msg [(x, Atom x, False, Nothing),(x, Atom x, False, Nothing)]
secGoal msg agents = Secret msg (map (\ x -> (x, Atom x, False, Nothing)) agents)

getAgentIds :: Types -> [Ident]
getAgentIds [] = []
getAgentIds ((Agent{},idents):xs)  = idents ++ getAgentIds xs
getAgentIds (_:xs) = getAgentIds xs 
 
getIdentsofMsg :: [Msg] -> [Ident]
getIdentsofMsg [] = []
getIdentsofMsg (x:xs) = case x of 
                        Atom a -> nub (a:getIdentsofMsg xs)
                        Comp _ _ msgs -> getIdentsofMsg (nub (msgs++xs))

-- function with TypeSignature Annotation (DO NOT DELETE, may be useful in the future)
--types2context c ((Function to, ids):xs) = let
--                                                 funsig = [x | (FunSign x) <- to]
--                                                 vb z = case funsig of
--                                                          [] -> VarBind (TFunction z ([TMsg],TMsg))
--                                                          fn -> let 
--                                                                  tn = map mapAtomicTypes (fst(head fn))
--                                                                  t = mapAtomicTypes (snd(head fn))
--                                                               in VarBind (TFunction z (tn,t))                 
--                                                 c1 = foldl (\x y -> appendBinding x y (vb y)) c ids
--                                          in types2context c1 xs

types2context :: Context -> Types -> Context
types2context c [] = c

types2context c ((Function _, ids):xs) = let
                                                 vb z = VarBind (TFunction z [TMsg])
                                                 c1 = foldl (\x y -> appendBinding x y (vb y)) c ids
                                         in types2context c1 xs

types2context c ((t,ids):xs) = let
                                    vb = VarBind (mapAtomicTypes t)
                                    c1 = foldl (\x y -> appendBinding x y vb) c ids
                               in types2context c1 xs

mappings2context :: Context -> Mappings -> Context
mappings2context c [] = c
mappings2context c ((id,ts,t):xs) = mappings2context c1 xs
                                 where
                                    c1 = appendBinding c id vb
                                    vb = VarBind (TMapping id (map mapAtomicTypes ts, mapAtomicTypes t)) 

formats2context :: Context -> Formats -> Context
formats2context c [] = c
formats2context c ((id,ts):xs) = formats2context c1 xs
                                 where
                                    c1 = appendBinding c id vb
                                    vb = VarBind (TFormat id (map mapAtomicTypes ts))  

actions2context :: Context -> Actions -> Context
actions2context c [] = c
actions2context c (x:xs) = let 
                                c1 = case x of
                                        CreateFresh (_, x) -> types2context c [x] 
                                        _ -> c
                           in actions2context c1 xs

mapAtomicTypes :: Type -> TermType
mapAtomicTypes t = case t of 
                Boolean _       -> TBoolean
                Number _        -> TNumber
                Agent{}         -> TAgent
                PublicKey _     -> TPublicKey
                PrivateKey _    -> TPrivateKey
                SymmetricKey _  -> TSymmetricKey
                SeqNumber _     -> TSeqNumber 
                Untyped _       -> TUntyped 
                Msg _           -> TMsg
                _               -> error ("Unhandled type " ++ show t)              
                
                
--typeCheckMsg :: Msg -> Context -> Context
--typeCheckMsg msg ctx = case typeof_ts (AMsg msg) ctx of
--                            _ -> ctx
--
--typeCheckAction :: Action -> Context -> Context
--typeCheckAction (Exchange (_,m1,m2,m3)) ctx = let
--                                                    c1 =  typeCheckMsg m1 ctx
--                                                    c2 =  case m2 of
--                                                                Nothing -> c1
--                                                                Just m -> typeCheckMsg m c2
--                                                    c3 =  case m3 of
--                                                                Nothing -> c2
--                                                                Just m -> typeCheckMsg m c2
--                                               in c3
--
--typeCheckAction (Let (_, m)) ctx = typeCheckMsg m ctx
--typeCheckAction (If (_,a1,a2)) ctx = typeCheckAction a2 c1
--                                        where
--                                          c1 = typeCheckAction a1 ctx
--typeCheckAction (Or (a1,a2)) ctx = typeCheckAction a2 c1
--                                        where
--                                          c1 = typeCheckAction a1 ctx
--typeCheckAction (CreateFresh _) ctx = ctx                
               
-- Operational Strands

sknowledge2context :: Context -> SKnowledge -> Context
sknowledge2context ctx [] = ctx
sknowledge2context ctx (x@(m1,m2):xs) = case m1 of
                                            Atom id -> let
                                                            t = typeof_ts (AMsg m2) ctx
                                                            c1 = appendBinding ctx id (VarBind t)
                                                       in sknowledge2context c1 xs
                                            _ -> error ("knowledge " ++ show x ++ " unhandled")                

showFullTypedStrand :: [Strand] -> Context -> String
showFullTypedStrand [] _ = ""
showFullTypedStrand (x:xs) ctx = showTypedStrand x ctx ++ "\n" ++ showFullTypedStrand xs ctx 

typecheckStrand :: Strand -> Context -> Context
typecheckStrand (Strand _ _ k r) ctx = let
                                            c1 = sknowledge2context ctx k
                                            c2 = typecheckStrand0 r c1 
                                       in c2

showTypedStrand :: Strand -> Context -> String
showTypedStrand os@(Strand n o k r) ctx = let
                                            c2 = typecheckStrand os ctx
                                          in "\n" ++ show c2 ++
                                             "\nPROTOCOL: " ++n ++ "\nOWNER: " ++ o ++ "\nKNOWLEDGE: " ++ show k ++ "\nREST:\n" ++ showTypedStrand0 r c2

typecheckStrand0 :: Strand0 -> Context -> Context
typecheckStrand0 (Snd _ _ rest) ctx = typecheckStrand0 rest ctx
typecheckStrand0 (Rcv (_,expr,_,_) (Atom id) rest) ctx = typecheckStrand0 rest c1
                                                                    where 
                                                                        vb = VarBind (typeof_ts (AMsg expr) ctx)
                                                                        c1 = appendBinding ctx id vb
typecheckStrand0 (IF _ tb fb rest) ctx = let
                                            c1 = typecheckStrand0 tb ctx
                                            c2 = typecheckStrand0 fb c1
                                         in typecheckStrand0 rest c2
typecheckStrand0 (OR lb rb rest) ctx = let
                                            c1 = typecheckStrand0 lb ctx
                                            c2 = typecheckStrand0 rb c1
                                       in  typecheckStrand0 rest c2
typecheckStrand0 (Add _ _ _ _  rest) ctx   = typecheckStrand0 rest ctx 
typecheckStrand0 (Delete _ rest) ctx       = typecheckStrand0 rest ctx
typecheckStrand0 (Query _ rest) ctx        = typecheckStrand0 rest ctx
typecheckStrand0 (Omar _ rest) ctx         = typecheckStrand0 rest ctx
typecheckStrand0 (Connects rest) ctx         = typecheckStrand0 rest ctx
typecheckStrand0 (How d (Atom id) m2 m3 rest) ctx  = typecheckStrand0 rest c1
                                                            where 
                                                                vb = VarBind (typeof_ts (ADes d m2 m3) ctx)
                                                                c1 = appendBinding ctx id vb
typecheckStrand0 (V v m1 m2 rest) ctx       = let 
                                                    term = AVer v m1 m2
                                                    t = typeof_ts term ctx
                                                  in
                                                  if t == TBoolean then typecheckStrand0 rest ctx else
                                                    error ("type of " ++ show term ++ " must be boolean")

typecheckStrand0 (Create _ _ _ rest) ctx = typecheckStrand0 rest ctx
typecheckStrand0 (Fresh t ms@(Atom id : _) m3 rest) ctx    = let 
                                                    vb = VarBind (mapAtomicTypes t)
                                                    c1 = appendBinding ctx id vb
                                               in if tail ms /= [] then typecheckStrand0 (Fresh t (tail ms) m3 rest) c1 else typecheckStrand0 rest c1

typecheckStrand0 (EOS _) ctx              = ctx
typecheckStrand0 str _ = error ("unexpected strand type: " ++ show str)

showTypedStrand0 :: Strand0 -> Context -> String
showTypedStrand0 (Snd det m rest) ctx = "\t\t  |---SND---> "++ show m ++ "/" ++ show (typeof_ts (AMsg m) ctx) ++ "/ " ++ "("++ show det ++")\n"++ showTypedStrand0 rest ctx
showTypedStrand0 (Rcv det m rest) ctx = "\t\t  |<--RCV---- "++ show m ++ "/" ++ show (typeof_ts (AMsg m) ctx) ++ "/ " ++ "("++ show det ++")\n"++ showTypedStrand0 rest ctx
showTypedStrand0 (IF fct tb fb rest) ctx     = "\t\t /\\   IF     "++ show fct ++ "\n\t True Branch\n\t" ++ showTypedStrand0 tb ctx ++"\t\t\t False Branch \t\t\t"++ showTypedStrand0 fb ctx ++"\n" ++ showTypedStrand0 rest ctx
showTypedStrand0 (OR lb rb rest) ctx         = "\t\t /\\   OR     "++"\n\t Left Branch\n\t" ++ showTypedStrand0 lb ctx ++"\t\t\t Right Branch \t\t\t"++ showTypedStrand0 rb ctx ++"\n" ++ showTypedStrand0 rest ctx
showTypedStrand0 (Add fct s m ms rest) ctx   = "\t\t  |    +      "++ show fct ++ "("++show m ++ show ms ++")["++ show s ++"]\n" ++ showTypedStrand0 rest ctx
showTypedStrand0 (Delete fct rest) ctx       = "\t\t  |    -      "++ show fct ++ showTypedStrand0 rest ctx
showTypedStrand0 (Query _ rest) ctx          = "\t\t  |    ?      "++ showTypedStrand0 rest ctx
showTypedStrand0 (Omar str rest) ctx         = "\t\t  |    omar   "++ str ++ showTypedStrand0 rest ctx
showTypedStrand0 (Connects rest) ctx         = showTypedStrand0 rest ctx
showTypedStrand0 (How d m1 m2 m3 rest) ctx   = "\t\t  |    macro  "++ show m1 ++ "/" ++ show (typeof_ts (AMsg m1) ctx) ++ "/ " ++ ":=" ++ show d ++ "("++ show m2 ++ ")\t["++ show m3 ++"]" ++ "\n" ++ showTypedStrand0 rest ctx
showTypedStrand0 (V v m1 m2 rest) ctx        = "\t\t  |    check  "++ show v  ++ "(" ++show m1 ++ ")==" ++ show m2 ++ "\n" ++ showTypedStrand0 rest ctx
showTypedStrand0 (Create m1 m2 m3 rest) ctx  = "\t\t  |    format "++ show m1 ++"=" ++ show m2 ++ "(" ++ show m3 ++ ")" ++ "\n" ++showTypedStrand0 rest ctx
showTypedStrand0 (Fresh t ms m3 rest) ctx    = "\t\t  |    Fresh  "++ show (mapAtomicTypes t) ++ " " ++ show (head ms) ++" ("++ show m3 ++ ")" ++ "\n" ++
                                                                       if tail ms /= [] then showTypedStrand0 (Fresh t (tail ms) m3 rest) ctx 
                                                                                        else showTypedStrand0 rest ctx
showTypedStrand0 (EOS sess) _                = "\t\t  - EOS       "++ sess

-- main

ttsOOS :: Protocol -> String
ttsOOS prot = let
                ctx = buildContext prot
                os = opStrands . generateStrands $prot  
              in showFullTypedStrand os ctx                          